:: ============================================================
:: Libmap.cmd(mappings library) - A Windows batch script helper for managing key:value mappings
::
:: The file format it understands is a subset of YAML (yaml.org)
::
::   e.g. key: Some value
::
:: Usage: libmap.cmd [GLOBAL_OPTIONS...] ACTION [ACTION_OPTIONS...] [PARAMETERS...]
::
:: Supported actions:
::   list           output the specified group of key, value pairs
::   get            retrieve the value associated with a key
::   set            map the given key to a value
::   delete         remove a pair from a group based on the given key
::   save           write the specified group of mappings to the given file
::   load           load mappings for group from given data file
::   clear          clear all the in memory mapping data for the given group
::   has_data       determine if the given group has any mappings
::
:: Metainfo accessors:
::   :xcode         display the message corresponding to the given status code
::   :about         display information about the library. Copyright, license, etc.
::   :version       return/display the version of the library
::
:: Global options:
::   -v             verbose, display all output and messages (info, error, etc.)
::
:: Copyright (c) 2010, Rudy Lattae
:: License: Simplified BSD License
:: ============================================================

@echo off
setlocal
shift


:: ============================================================
:: Initialise
:: ============================================================
:lib.init -- global constants, variables and config options
    :: Info -- name, version, copyright, etc.    
    set "LIBMAP_INFO_NAME=libmap.cmd"
    set "LIBMAP_INFO_SHORTNAME=libmap"
    set "LIBMAP_INFO_LONGNAME=Libmap.cmd(mappings library)"
    set "LIBMAP_INFO_VERSION=0.9.12"
    set "LIBMAP_INFO_SUMMARY=A Windows batch script helper for managing key:value mappings"
    set "LIBMAP_INFO_COPYRIGHT=Copyright (c) 2010, Rudy Lattae"
    set "LIBMAP_INFO_LICENSE=Simplified BSD License"

    :: Configuration options 
    if "%LIBMAP_OPT_VERBOSE%"=="" set LIBMAP_OPT_VERBOSE="False"

    :: Status (x) codes
    set LIBMAP_X_INVALID_GLOBAL_OPTION=1
    set LIBMAP_X_ACTION_REQUIRED=2
    set LIBMAP_X_INVALID_ACTION=3
    set LIBMAP_X_INVALID_ARGUMENTS=4
    set LIBMAP_X_FILE_NOT_FOUND=5


:: ============================================================
:: Front controller
:: ============================================================

:lib.main
    :: capture any global options
    :lib.main._parse_global_options
    set "libmap__opt=%0"
    if "%libmap__opt:~0,1%"=="-" (
        if "%libmap__opt%"=="-v" (set LIBMAP_OPT_VERBOSE="True" & shift & goto lib.main._parse_global_options)
        if %LIBMAP_OPT_VERBOSE%=="True" call :msg.invalid_global_option %libmap__opt%, "%LIBMAP_INFO_NAME%: ERROR: "
        exit /b %LIBMAP_X_INVALID_GLOBAL_OPTION%
    )
    
    :: make sure we have an action, there is no need to proceed otherwise
    if "%0"=="" (
        if %LIBMAP_OPT_VERBOSE%=="True" call :msg.action_required "%LIBMAP_INFO_NAME%: ERROR: "
        exit /b %LIBMAP_X_ACTION_REQUIRED%
    )
    
    :: attempt to handle the requested action
    if "%0"=="list" goto do.list
    if "%0"=="get" goto do.get
    if "%0"=="set" goto do.set
    if "%0"=="delete" goto do.delete
    if "%0"=="save" goto do.save
    if "%0"=="load" goto do.load
    if "%0"=="clear" goto do.clear
    if "%0"=="has_data" goto do.has_data
    
    :: process any requests for library meta data
    if "%0"==":xcode" goto meta.xcode
    if "%0"==":about" goto meta.about
    if "%0"==":version" goto meta.version
    
    :: if we get here, we know the action requested is not supported - leave!
    if %LIBMAP_OPT_VERBOSE%=="True" call :msg.invalid_action %0 "%LIBMAP_INFO_NAME%: ERROR: "
    
    exit /b %LIBMAP_X_INVALID_ACTION%


:: ============================================================
:: Action handlers
:: ============================================================

:do.list GROUP -- output a list of the mappings in the given group
:: GROUP                    : in
:: -separator ARG, -sep ARG : default=" => " -- character/string to use to seperate each key-value pair in the output
:: -delims ARG              : default="__=" -- charaters to use as delimeters for parts of "namespaces" group variable
:: -tokens ARG              : default=3* -- number of tokens to expect in group variable
    set "_separator= =^> "
    set "_delims=__="
    set "_tokens=3*"

    :do.list._parse_options
    set "__opt=%1"
    if "%__opt:~0,1%"=="-" (
        if "%__opt%"=="-separator" (set "_separator=%~2" & shift /1 & shift /1 & goto do.list._parse_options)
        if "%__opt%"=="-sep" (set "_separator=%~2" & shift /1 & shift /1 & goto do.list._parse_options)
        if "%__opt%"=="-delims" (set "_delims=%~2" & shift /1 & shift /1 & goto do.list._parse_options)
        if "%__opt%"=="-tokens" (set "_tokens=%~2" & shift /1 & shift /1 & goto do.list._parse_options)
        exit /b %LIBMAP_X_INVALID_GLOBAL_OPTION%
    )
    if not "%1"=="" (set "_group=%~1") else (exit /b %LIBMAP_X_INVALID_ARGUMENTS%)
    
    (endlocal
        echo.
        for /f "tokens=%_tokens% delims=%_delims%" %%a in ('set %_group%') do (
            if %%a NEQ "" echo.%%a%_separator%%%b
        )
    )
    exit /b 0

:do.get GROUP KEY [RESULT] -- retrieve the value associated with a given key
:: GROUP       : in
:: KEY         : in
:: [RESULT]    : out, optional
:: -delims ARG : default="__=" -- charaters to use as delimeters for parts of "namespaces" group variable
:: -tokens ARG : default=3* -- number of tokens to expect in group variable
    set "_delims=__="
    set "_tokens=3*"
    set "_result="
    
    :do.get._parse_options
    set "__opt=%1"
    set "__opt_delims_key=delims"
    if "%__opt:~0,1%"=="-" (
        if "%__opt%"=="-delims" (set "_delims=%~2" & shift /1 & shift /1 & goto do.get._parse_options)
        if "%__opt%"=="-tokens" (set "_tokens=%~2" & shift /1 & shift /1 & goto do.get._parse_options)
        exit /b %LIBMAP_X_INVALID_GLOBAL_OPTION%
    )

    if not "%1"=="" (set "_group=%~1") else (exit /b %LIBMAP_X_INVALID_ARGUMENTS%)
    if not "%2"=="" (set "_key=%~2") else (exit /b %LIBMAP_X_INVALID_ARGUMENTS%)
    
    for /f "tokens=%_tokens% delims=%_delims%" %%a in ('set %_group% ') do (
        if %%a==%_key% (set "_result=%%b" & goto do.get._done)
    )
    
    :do.get._done
    (endlocal
        if not "%~3"=="" (set %~3=%_result%) else (echo.%_result%)
    )
    exit /b 0

:do.set GROUP KEY VALUE -- map a key to a given value
:: GROUP : in
:: KEY   : in
:: VALUE : in
    if not "%1"=="" (set "_group=%~1") else (exit /b %LIBMAP_X_INVALID_ARGUMENTS%)
    if not "%2"=="" (set "_key=%~2") else (exit /b %LIBMAP_X_INVALID_ARGUMENTS%)
    if not "%~3"=="" (set "_value=%~3") else (exit /b %LIBMAP_X_INVALID_ARGUMENTS%)

    (endlocal
        set "%_group%_%_key%=%_value%"
    )
    exit /b 0

:do.delete GROUP KEY -- remove a reference entry from the given map group
:: GROUP : in
:: KEY   : in
    if not "%1"=="" (set "_group=%~1") else (exit /b %LIBMAP_X_INVALID_ARGUMENTS%)
    if not "%2"=="" (set "_key=%~2") else (exit /b %LIBMAP_X_INVALID_ARGUMENTS%)
    
    (endlocal
        set %_group%_%_key%=
    )
    exit /b 0

:do.save GROUP DATA_FILE_PATH [CREATE_FILE="True"]) -- write the specified group of mappings to the given file
:: GROUP            : in
:: DATA_FILE_PATH   : in
:: -create-file ARG : default="True" -- flag ("True"/"False") if file should be created in case it does not exist yet
    set _create_file="True"
    
    :do.get._parse_options
    set "__opt=%1"
    set "__opt_delims_key=delims"
    if "%__opt:~0,1%"=="-" (
        if "%__opt%"=="-create-file" (set "_create_file=%~2" & shift /1 & shift /1 & goto do.get._parse_options)
        exit /b %LIBMAP_X_INVALID_GLOBAL_OPTION%
    )
    if not "%1"=="" (set "_group=%~1") else (exit /b %LIBMAP_X_INVALID_ARGUMENTS%)
    if not "%2"=="" (set "_data_file_path=%~2") else (exit /b %LIBMAP_X_INVALID_ARGUMENTS%)

    if not exist "%_data_file_path%" (
        if %_create_file%=="True" (call :util.create_file %_data_file_path% & goto :do.save._file_exists)
        if %LIBMAP_OPT_VERBOSE%=="True" (call :msg.file_not_found %_data_file_path%, "%LIBMAP_INFO_NAME%: save: ERROR: ")
        exit /b %LIBMAP_X_FILE_NOT_FOUND%
    )
    
    :do.save._file_exists
    call :do.list -sep ": " %_group%>"%_data_file_path%"
    endlocal
    exit /b 0

:do.load GROUP DATA_FILE_PATH -- load mappings for group from given data file
:: GROUP          : in
:: DATA_FILE_PATH : in
    if not "%1"=="" (set "_group=%~1") else (exit /b %LIBMAP_X_INVALID_ARGUMENTS%)
    if not "%2"=="" (set "_data_file_path=%~2") else (exit /b %LIBMAP_X_INVALID_ARGUMENTS%)

    if not exist "%_data_file_path%" (
        if %LIBMAP_OPT_VERBOSE%=="True" call :msg.file_not_found %_data_file_path%, "%LIBMAP_INFO_NAME%: load: ERROR: "
        exit /b %LIBMAP_X_FILE_NOT_FOUND%
    )

    (endlocal    
        for /f "usebackq tokens=1,* delims=: " %%a in ("%_data_file_path%") do (
            set "%_group%_%%a=%%b"
        )
    )
    exit /b 0

:do.clear GROUP -- clear all the in memory mapping data for the given group
:: GROUP                    : in
:: -separator ARG, -sep ARG : default=" => " -- character/string to use to seperate each key-value pair
:: -tokens ARG              : default=3* -- number of tokens to expect in group variable
:: -delims ARG              : default="__=" -- charaters to use as delimeters for parts of "namespaces" group variable
    set "_tokens=3*"
    set "_delims=__="
    
    :do.clear._parse_options
    set "__opt=%1"
    if "%__opt:~0,1%"=="-" (
        if "%__opt%"=="-tokens" (set "_tokens=%~2" & shift /1 & shift /1 & goto do.clear._parse_options)
        if "%__opt%"=="-delims" (set "_delims=%~2" & shift /1 & shift /1 & goto do.clear._parse_options)
        exit /b %LIBMAP_X_INVALID_GLOBAL_OPTION%
    )

    if not "%1"=="" (set "_group=%~1") else (exit /b %LIBMAP_X_INVALID_ARGUMENTS%)

    (endlocal
        for /f "tokens=%_tokens% delims=%_delims%" %%a in ('set %_group%') do (
            set "%_group%_%%a="
        )
    )
    exit /b 0

:do.has_data GROUP [RESULT] -- determine if the given map group has mappings
:: GROUP                    : in
:: RESULT                   : out, optional, boolean, ("True", "False")
:: -tokens ARG              : default=3* -- number of tokens to expect in group variable
:: -delims ARG              : default="__=" -- charaters to use as delimeters for parts of "namespaces" group variable
    set "_tokens=3*"
    set "_delims=__="
    set _result="False"
    
    :do.has_data._parse_options
    set "__opt=%1"
    if "%__opt:~0,1%"=="-" (
        if "%__opt%"=="-tokens" (set "_tokens=%~2" & shift /1 & shift /1 & goto do.has_data._parse_options)
        if "%__opt%"=="-delims" (set "_delims=%~2" & shift /1 & shift /1 & goto do.has_data._parse_options)
        exit /b %LIBMAP_X_INVALID_GLOBAL_OPTION%
    )
    
    if not "%1"=="" (set "_group=%~1") else (exit /b %LIBMAP_X_INVALID_ARGUMENTS%)
    
    (endlocal
        for /f "tokens=%_tokens% delims=%_delims%" %%a in ('set %_group%') do (
            if %%a NEQ "" (set _result="True" & goto :do.has_data._break)
        )
        
        :do.has_data._break
        if not "%~2"=="" (set %~2=%_result%) else (echo.%_result%)
    )
    exit /b 0


:: ============================================================
:: Metainfo accessors
:: ============================================================

:meta.xcode STATUS_CODE [RESULT] -- display the message corresponding to the given status code
:: -- STATUS_CODE : in
:: -- RESULT      : out, optional
    if not "%1"=="" (set "_status_code=%~1") else (goto :eof)
    set "_result="
    
    if "%_status_code%"=="%LIBMAP_X_INVALID_GLOBAL_OPTION%" (
        set "_result=Invalid global option" & goto meta.xcode._done
    )
    if "%_status_code%"=="%LIBMAP_X_ACTION_REQUIRED%" (
        set "_result=Action required" & goto meta.xcode._done
    )
    if "%_status_code%"=="%LIBMAP_X_INVALID_ACTION%" (
        set "_result=Invalid action" & goto meta.xcode._done
    )
    if "%_status_code%"=="%LIBMAP_X_INVALID_ARGUMENTS%" (
        set "_result=Invalid arguments" & goto meta.xcode._done
    )
    if "%_status_code%"=="%LIBMAP_X_FILE_NOT_FOUND%" (
        set "_result=File not found" & goto meta.xcode._done
    )
    
    :meta.xcode._done
    (endlocal
        if not "%~2"=="" (set %~2="%LIBMAP_INFO_NAME%: %_result%") else (echo.%LIBMAP_INFO_NAME%: %_result%)
    )
    goto :eof

:meta.about -- display information about the library. Copyright, license, etc.
    echo.%LIBMAP_INFO_LONGNAME% (version %LIBMAP_INFO_VERSION%)
    echo.
    echo.  %LIBMAP_INFO_SUMMARY%
    echo.
    echo.  %LIBMAP_INFO_COPYRIGHT%
    echo.  %LIBMAP_INFO_LICENSE%
    goto :eof

:meta.version [RESULT] -- return/display the version of the library
:: -- RESULT : out, optional
    if "%~1"=="" (echo.%LIBMAP_INFO_VERSION% & goto :eof)
    (endlocal
        set "%~1=%LIBMAP_INFO_VERSION%"
    )
    goto :eof


:: ============================================================
:: Messages
:: ============================================================

:msg.invalid_global_option (option, [meta])
    setlocal
    set "_option=%~1"
    set "_meta=%~2"
    echo.%_meta%Invalid global option "%_option%"
    endlocal
    goto :eof

:msg.action_required ([meta])
    setlocal    
    set "_meta=%~1"
    echo.%_meta%Action required.
    endlocal
    goto :eof

:msg.invalid_action (action, [meta])
    setlocal
    set "_action=%~1"
    set "_meta=%~2"
    echo.%_meta%Invalid action "%_action%"
    endlocal
    goto :eof

:msg.file_not_found (file_path, [meta])
:: file_path : in
    setlocal
    set "_file_path=%~1"
    set "_meta=%~2"
    echo.%_meta%Cannot locate the file: %_file_path%
    endlocal
    goto :eof
    
    
:: ============================================================
:: Utilities
:: ============================================================
   
:util.create_file (file_path) -- create a file if it does not yet exist
:: file_path : in
    setlocal
    if not "%1"=="" (set "_file_path=%~1") else (goto :eof)
    set "_dir=%~dp1"

    if not exist "%_dir%" mkdir "%_dir%"
    if not exist "%_file_path%" echo.>"%_file_path%"
    endlocal
    goto :eof
